TYPE=OVSIntPort
OVS_BRIDGE=br-int
DEVICETYPE=ovs
ONBOOT=yes
BOOTPROTO=static
IPV6_AUTOCONF=no
DEVICE={{ mgmt_port_dev }}
IPADDR={{ mgmt_port_ip }}
NETMASK={{ mgmt_port_netmask }}
NM_CONTROLLED=no
MACADDR={{ mgmt_port_mac }}
MTU={{ mgmt_port_mtu }}
OVS_EXTRA="-- set Interface {{ mgmt_port_dev }} external-ids:iface-status=active \
           -- set Interface {{ mgmt_port_dev }} external-ids:attached-mac={{ mgmt_port_mac }} \
           -- set Interface {{ mgmt_port_dev }} external-ids:iface-id={{mgmt_port_id }} \
           -- set Interface {{ mgmt_port_dev }} external-ids:skip_cleanup=true \
           -- set Interface {{ mgmt_port_dev }} mac=\"{{ mgmt_port_mac }}\" \
           -- set Interface {{ mgmt_port_dev }} other-config:hwaddr={{ mgmt_port_mac }}"

